"""
The local queue transport roughly mimics a normal TCP transport, but it sends
and receives messages via two :class:`asyncio.Queue` instances.

Its purpose is to aid the development and debugging of complex networking
algorithms and distributed or multi-agent systems.  In contrast to normal
network transports, messages send via the :class:`LocalQueueTransport` will
always arrive in a deterministic order [#]_.

This transport does *not* work across multiple processes and is *not* thread
safe, so it should only be used within a single thread and process.

The easiest way to use it is to create a :class:`LocalQueue` instance via the
:func:`get_queue()` function and pass it to
:func:`aiomas.channel.start_server()`/:func:`aiomas.channel.open_connection()`
or :meth:`aiomas.agent.Container.create()` as *addr* argument.

.. [#] Actually, message sent via a single TCP connection also arrive at a
       deterministic order (this is a property of the TCP/IP protocol).  So
       the LocalQueue transport won't give you any benefits in this case.

       However, if you have multiple connections to the same server and send
       message through them in parallel, it's no longer deterministic in which
       order the messages arrive from the different connections.  In this case,
       the LocalQueue transport can help you.

"""
import asyncio


__all__ = [
    'get_queue', 'clear_queue_cache',
    'create_connection', 'create_server',
    'LocalQueue', 'LocalQueueServer', 'LocalQueueTransport',
]


CLOSE_QUEUE = object()
_QUEUES = {}


def get_queue(queue_id):
    """Return a :class:`~aiomas.local_queue.LocalQueue` instance for the given
    *queue_id*.

    If no instance is cached yet, create a new one.

    Queue IDs must be strings and must not contain the ``/`` character.  Raise
    a :exc:`ValueError` if these rules are violated.

    """
    if not isinstance(queue_id, str):
        raise ValueError('Queue ID must be "str" not "{}"'
                         .format(type(queue_id)))

    if '/' in queue_id:
        raise ValueError('"/" not allowed in queue ID ({!r})'.format(queue_id))

    return _QUEUES.setdefault(queue_id, LocalQueue(queue_id))


def clear_queue_cache():
    """Clear the global queue cache."""
    _QUEUES.clear()


async def create_connection(protocol_factory, lq, *, loop=None, **kwds):
    """Connect to a :class:`LocalQueue` *lq* and return a
    *(transport, protocol)* pair.

    The *protocol_factory* must be a callable returning a `protocol
    <https://docs.python.org/3/library/asyncio-protocol.html>`_ instance.

    Before a connection to *lq* can be made, a server must
    be started for this instance (see :func:`create_server()`).

    """
    if loop is None:
        loop = asyncio.get_event_loop()

    client_server_q = asyncio.Queue(loop=loop)
    server_client_q = asyncio.Queue(loop=loop)

    lq.new_connection(sendq=server_client_q, recvq=client_server_q)

    p = protocol_factory()
    t = LocalQueueTransport(lq, sendq=client_server_q, recvq=server_client_q,
                            protocol=p, loop=loop)
    p.connection_made(t)
    return (t, p)


async def create_server(protocol_factory, lq, **kwds):
    """Create a :class:`LocalQueue` server bound to *lq*.

    The *protocol_factory* must be a callable returning a `protocol
    <https://docs.python.org/3/library/asyncio-protocol.html>`_ instance.

    Return a :class:`LocalQueueServer` instance.  That instance is also set
    as :attr:`~LocalQueue.server` for *lq*.

    """
    return LocalQueueServer(protocol_factory, lq)


class LocalQueue:
    """An instance of this class serves as transport description when creating
    a server or connection.

    The functions :func:`create_server()` and :func:`create_connection()` both
    require an instance of this class.  Alternatively, instances of this class
    can be passed as *addr* argument to :func:`aiomas.channel.start_server()`
    and :func:`aiomas.channel.open_connection()`

    A server needs to be started before any connections can be made.

    """
    def __init__(self, queue_id):
        self._queue_id = queue_id
        self._server = None

    def __repr__(self):
        return '<{}.{} object {!r} at 0x{:x}>'.format(
            self.__class__.__module__,
            self.__class__.__name__,
            self._queue_id,
            id(self),
        )

    def __str__(self):
        return '{}({!r})'.format(self.__class__.__name__, self._queue_id)

    @property
    def queue_id(self):
        """The queue's ID."""
        return self._queue_id

    @property
    def server(self):
        """The :class:`LocalQueueServer` instance that was bound to this
        instance or ``None`` if no server has yet been started.

        """
        return self._server

    def set_server(self, server):
        """Set a :class:`LocalQueueServer` instance.

        Raise a :exc:`RuntimeError` if a server has already been bound to
        this instance.

        This method is called by :func:`create_server()`.

        """
        if self._server is not None:
            raise RuntimeError('Server is already set.')

        self._server = server

    def unset_server(self):
        """Unset the server from this instance.

        This method is called when the server is closed (see
        :class:`LocalQueueServer.close()`).

        """
        self._server = None

    def new_connection(self, sendq, recvq, loop=None):
        """Create a connection endpoint on the server side.

        This method is called by :func:`create_connection()`.

        *sendq* and *recvq* are the queues used for sending and receiving
        messages to and from the client.

        """
        if self._server is None:
            raise ConnectionRefusedError('No server started for this local '
                                         'queue instance.')
        if loop is None:
            loop = asyncio.get_event_loop()
        self._server.new_connection(sendq, recvq, loop)


class LocalQueueServer(asyncio.AbstractServer):
    """Implements ``asyncio.events.AbstractServer``.  An instance of this class
    is returned by :func:`create_server()`.

    *lq* is the :class:`LocalQueue` instance that this server was bound to.

    *protocol_factory* is a callable that is called for each new client
    connection in order to create a new protocol instance.

    """
    def __init__(self, protocol_factory, lq):
        self._protocol_factory = protocol_factory
        self._lq = lq
        lq.set_server(self)
        self._transports = []

    @property
    def lq(self):
        """The :class:`LocalQueue` the server is bound to."""
        return self._lq

    def new_connection(self, sendq, recvq, loop):
        """Create a new protocol and transport instance.

        Call the *protocol factory*, create a new :class:`LocalQueueTransport`
        with *sendq* and *recvq* and wire them together.

        Called by :func:`create_connection()` via
        :meth:`LocalQueue.new_connection()`.

        """
        p = self._protocol_factory()
        t = LocalQueueTransport(self._lq, sendq, recvq, p, loop)
        self._transports.append(t)
        p.connection_made(t)

    def close(self):
        """Close the server and unset this instance from the associated
        :class:`LocalQueue` instance.

        """
        self._lq.unset_server()

    async def wait_closed(self):
        """Immediately return (there's nothing to wait for).

        """
        if self._transports:
            await asyncio.gather(
                *[t._task_recv for t in self._transports],
                return_exceptions=True,
                loop=self._transports[0]._loop,
            )


class LocalQueueTransport(asyncio.Transport):
    """Implements ``asyncio.transports.Transport``.

    A *LocalQueueTransport* has two asynchronous queues (instances of
    :class:`asyncio.Queue`) – one for sending messages to the other side
    and one for receiving messages from it.

    """
    def __init__(self, lq, sendq, recvq, protocol, loop):
        super().__init__(extra={'peername': str(lq)})
        self._sendq = sendq
        self._recvq = recvq
        self._protocol = protocol
        self._loop = loop
        self._wait_read = None
        self._task_recv = loop.create_task(self._recv())

    async def _recv(self):
        try:
            while self._recvq is not None:
                if self._wait_read is not None:
                    await self._wait_read
                data = await self._recvq.get()
                if data is CLOSE_QUEUE:
                    self.close()
                    break
                self._protocol.data_received(data)
        except asyncio.CancelledError:
            return

    def close(self):
        """Close the transport.

        Buffered data will be flushed asynchronously.  No more data will be
        received.  After all buffered data is flushed, the protocol's
        ``connection_lost()`` method will (eventually) be called with ``None``
        as its argument.

        """
        if self._sendq is not None:
            sendq = self._sendq
            self._task_recv.cancel()
            self._sendq = None
            self._recvq = None
            sendq.put_nowait(CLOSE_QUEUE)
            self._protocol.connection_lost(None)
            return self._task_recv

    def write(self, data):
        """Write some data bytes to the transport.

        This does not block; it buffers the data and arranges for it to be sent
        out asynchronously.

        """
        self._sendq.put_nowait(data)

    def can_write_eof(self):
        """Return ``False``.  This transport does not support ``write_eof()``.
        """
        return False

    def abort(self):
        """Close the transport immediately.

        Buffered data will be lost.  No more data will be received.  The
        protocol's ``connection_lost()`` method will (eventually) be called
        with ``None`` as its argument.

        """
        self.close()



"""
Exception types used by *aiomas*.

"""


class AiomasError(Exception):
    """Base class for all exceptions defined by aiomas."""
    pass


class RemoteException(AiomasError):
    """Wraps a traceback of an exception on the other side of a channel.

    *origin* is the remote peername.

    *remote_traceback* is the remote exception's traceback.

    """
    def __init__(self, origin, remote_traceback):
        super().__init__(origin, remote_traceback)
        self.origin = origin  #: Peername (producer of the exception)
        self.remote_traceback = remote_traceback  #: Original traceback

    def __str__(self):
        return 'Origin: {}\n{}'.format(self.origin, self.remote_traceback)


class SerializationError(Exception):
    """Raised when an object cannot be serialized."""



import os
import sys

on_rtd = os.environ.get('READTHEDOCS', None) == 'True'

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, os.path.abspath(os.path.join('..', 'src')))

# -- General configuration ------------------------------------------------

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.intersphinx',
    'sphinx.ext.viewcode',
]

templates_path = ['_templates']

source_suffix = '.rst'

master_doc = 'index'

authors = ['Stefan Scherfke']
project = 'aiomas'
copyright = '2014, {}'.format(', '.join(authors))

version = '2.0'
release = '2.0.1'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build']

pygments_style = 'sphinx'


# -- Options for HTML output ----------------------------------------------

if on_rtd:
    html_theme = 'default'
else:
    html_theme = 'alabaster'

html_static_path = ['_static']

# Output file base name for HTML help builder.
htmlhelp_basename = 'aiomasdoc'


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    'papersize': 'a4paper',
    'pointsize': '11pt',
    # 'preamble': '',
}

latex_documents = [
    (master_doc, 'aiomas.tex', 'aiomas Documentation',
     ''.join(authors), 'manual'),
]


# -- Options for manual page output ---------------------------------------

man_pages = [
    (master_doc, 'aiomas', 'aiomas Documentation',
     authors, 1)
]


# -- Options for Texinfo output -------------------------------------------

texinfo_documents = [
    (master_doc, 'aiomas', 'aiomas Documentation',
     ''.join(authors), 'aiomas', 'One line description of project.',
     'Miscellaneous'),
]


# Intersphinx
intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
    'arrow': ('https://arrow.readthedocs.io/en/latest/', None),
}

# Autodoc
autodoc_member_order = 'bysource'



from . import (
    agent,
    channel,
    clocks,
    codecs,
    exceptions,
    local_queue,
    rpc,
    subproc,
    util,
)  # flake8: NOQA

from .agent import Container, Agent, SSLCerts
from .codecs import JSON, MsgPack, MsgPackBlosc, serializable
from .clocks import AsyncioClock, ExternalClock
from .exceptions import AiomasError, RemoteException
from .local_queue import get_queue
from .rpc import expose
from .util import (
    run,
    make_ssl_server_context,
    make_ssl_client_context,
)

__all__ = [
    # Decorators
    'expose', 'serializable',
    # Functions
    'get_queue',
    'run', 'make_ssl_server_context',
    'make_ssl_client_context',
    # Exceptions
    'AiomasError', 'RemoteException',
    # Classes
    'Container', 'Agent', 'SSLCerts',
    'JSON', 'MsgPack', 'MsgPackBlosc',
    'AsyncioClock', 'ExternalClock',
]
__version__ = '2.0.1'



"""
This module implements and asyncio :class:`asyncio.Protocol` protocol for a
request-reply :class:`Channel`.

"""
import asyncio
import collections
import itertools
import logging
import struct
import time
import traceback

from . import codecs
from . import local_queue
from .exceptions import RemoteException


__all__ = [
    'REQUEST', 'RESULT', 'EXCEPTION', 'DEFAULT_CODEC',
    'open_connection', 'start_server',
    'Header', 'Request', 'Channel', 'ChannelProtocol',
]


# Message types
REQUEST = 0
RESULT = 1
EXCEPTION = 2

DEFAULT_CODEC = codecs.JSON

Header = struct.Struct('!L')

logger = logging.getLogger(__name__)


async def open_connection(addr, *, loop=None, codec=None, extra_serializers=(),
                          timeout=0, **kwds):
    """Return a :class:`Channel` connected to *addr*.

    This is a convenience wrapper for
    :meth:`asyncio.AbstractEventLoop.create_connection()`,
    :meth:`asyncio.AbstractEventLoop.create_unix_connection()`, and
    :func:`aiomas.local_queue.create_connection()`.

    If *addr* is a tuple ``(host, port)``, a TCP connection will be created.
    If *addr* is a string, it should be a path name pointing to the unix domain
    socket to connect to.
    If *addr* is a :mod:`aiomas.local_queue` instance, a *LocalQueue*
    connection will be created.

    You can optionally provide the event *loop* to use.

    By default, the :class:`~aiomas.codecs.JSON` *codec* is used.  You
    can override this by passing any subclass of :class:`aiomas.codecs.Codec`
    as *codec*.

    You can also pass a list of *extra_serializers* for the codec.  The list
    entires need to be callables that return a tuple with the arguments for
    :meth:`~aiomas.codecs.Codec.add_serializer()`.

    With a *timeout* of 0 (the default), there will only be one connection
    attempt before an error is raised (:exc:`ConnectionRefusedError` for TCP
    sockets and LocalQueue, :exc:`FileNotFoundError` for Unix domain sockets).
    If you set *timeout* to a number > 0 or ``None``, this function will try to
    connect repeatedly for at most that many seconds (or indefinitely) before
    an error is raised.  Use this if you need to start the client before the
    server.

    The remaining keyword argumens *kwds* are forwarded to
    :meth:`asyncio.AbstractEventLoop.create_connection()` and
    :meth:`asyncio.AbstractEventLoop.create_unix_connection` respectively.

    This function is a `coroutine
    <https://docs.python.org/3/library/asyncio-task.html#coroutine>`_.

    """
    if loop is None:
        loop = asyncio.get_event_loop()

    if timeout is None:
        timeout = float('inf')

    codec = codec() if codec else DEFAULT_CODEC()
    for s in extra_serializers:
        codec.add_serializer(*s())

    def factory():
        return ChannelProtocol(codec, loop=loop)

    # Get the appropriate connect function and its arguments:
    if isinstance(addr, tuple):
        connect = loop.create_connection
        args = (factory,) + addr
        kwargs = kwds
        # If you connect to "localhost", some systems try to connect to
        # 127.0.0.1 and ::1 resulting in an OSError(Multiple errors occured),
        # that wraps two ConnectionRefusedErrors:
        ConnectionRefused = (ConnectionRefusedError, OSError)
    elif isinstance(addr, str):
        connect = loop.create_unix_connection
        args = (factory, addr)
        kwargs = kwds
        ConnectionRefused = FileNotFoundError
    elif isinstance(addr, local_queue.LocalQueue):
        connect = local_queue.create_connection
        args = (factory, addr)
        kwargs = dict(kwds, loop=loop)
        ConnectionRefused = ConnectionRefusedError
    else:
        raise ValueError('Unknown address type: %s' % addr)

    _t, p = await _try_connect(connect, args, kwargs, ConnectionRefused,
                               timeout)

    return p.channel


async def start_server(addr, client_connected_cb, *,
                       loop=None, codec=None, extra_serializers=(), **kwds):
    """Start a server listening on *addr* and call *client_connected_cb*
    for every client connecting to it.

    This function is a convenience wrapper for
    :meth:`asyncio.AbstractEventLoop.create_server()`,
    :meth:`asyncio.AbstractEventLoop.create_unix_server()`, and
    :func:`aiomas.local_queue.create_server()`.

    If *addr* is a tuple ``(host, port)``, a TCP socket will be created.
    If *addr* is a string, a unix domain socket at this path will be created.
    If *addr* is a :mod:`aiomas.local_queue` instance, a *LocalQueue* server
    will be created.

    The single argument of the callable *client_connected_cb* is a new instance
    of :class:`Channel`.

    You can optionally provide the event *loop* to use.

    By default, the :class:`~aiomas.codecs.JSON` *codec* is used.  You can
    override this by passing any subclass of :class:`aiomas.codecs.Codec` as
    *codec*.

    You can also pass a list of *extra_serializers* for the codec.  The list
    entires need to be callables that return a tuple with the arguments for
    :meth:`~aiomas.codecs.Codec.add_serializer()`.

    The remaining keyword argumens *kwds* are forwarded to
    :meth:`asyncio.AbstractEventLoop.create_server()` and
    :meth:`asyncio.AbstractEventLoop.create_unix_server` respectively.

    This function is a `coroutine
    <https://docs.python.org/3/library/asyncio-task.html#coroutine>`_.

    """
    if loop is None:
        loop = asyncio.get_event_loop()

    if not codec:
        codec = DEFAULT_CODEC

    def factory():
        c = codec()
        for s in extra_serializers:
            c.add_serializer(*s())
        return ChannelProtocol(c, client_connected_cb, loop=loop)

    if isinstance(addr, tuple):
        return await loop.create_server(factory, *addr, **kwds)
    elif isinstance(addr, str):
        return await loop.create_unix_server(factory, addr, **kwds)
    elif isinstance(addr, local_queue.LocalQueue):
        return await local_queue.create_server(factory, addr, **kwds)
    else:
        raise ValueError('Unknown address type: %s' % addr)


async def _try_connect(connect, args, kwargs, ConnectionRefused, timeout):
    """Try to establish a connection and return the *(transport, protocol)*
    pair if successful.

    Repeatedly call *connect(\\*args, \\*\\*kwargs)*.  If not
    connection can be opened for *timeout* seconds, re-raise the connection
    error from *connect* (which is of type *ConnectionRefused*).

    """
    retry_interval = 0.5
    start_time = time.monotonic()
    while True:
        try:
            t, p = await connect(*args, **kwargs)
            break
        except ConnectionRefused:
            if (time.monotonic() - start_time) > timeout:
                raise

            logger.debug('Could not connect.  Retrying in %.2fs ...',
                         retry_interval)
            await asyncio.sleep(retry_interval)
            retry_interval = min(30, 1.5 * retry_interval)  # Wait at most 30s

    return t, p


def _get_checked_msg_len(msg):
    """Return the number of bytes/length of *msg*.

    Raise a exc:`ValueError` if the length is to large for :class:`Header`.

    """
    msg_len = len(msg)
    try:
        len_bytes = Header.pack(msg_len)
    except struct.error:
        max_size = 2 ** (Header.size * 8)
        raise ValueError('Serialized message is too long (%d).  Maximum '
                         'length is %d.' % (msg_len, max_size)) from None

    return len_bytes


class ChannelProtocol(asyncio.Protocol):
    """Asyncio :class:`asyncio.Protocol` which connects the low level transport
    with the high level :class:`Channel` API.

    The *codec* is used to (de)serialize messages.  It should be a sub-class of
    :class:`aiomas.codecs.Codec`.

    Optionally you can also pass a function/coroutine *client_connected_cb*
    that will be executed when a new connection is made (see
    :func:`start_server()`).

    """
    def __init__(self, codec, client_connected_cb=None, *, loop):
        super().__init__()
        self.codec = codec
        self.transport = None
        self.channel = None
        self._client_connected_cb = client_connected_cb
        self._loop = loop
        self._buffer = bytearray()
        self._read_size = None

        # For flow control
        self._paused = False
        self._drain_waiter = None
        self._connection_lost = None
        self._out_msgs = asyncio.Queue()
        self._task_process_out_msgs = None

    def connection_made(self, transport):
        """Create a new :class:`Channel` instance for a new connection.

        Also call the *client_connected_cb* if one was passed to this class.

        """
        self._task_process_out_msgs = self._loop.create_task(
            self._process_out_msgs())

        self.transport = transport
        self.channel = Channel(self, self.codec, transport, loop=self._loop)

        if self._client_connected_cb is not None:
            res = self._client_connected_cb(self.channel)
            if asyncio.iscoroutine(res):
                self._loop.create_task(res)

    def connection_lost(self, exc):
        """Set a :exc:`ConnectionError` to the :class:`Channel` to indicate
        that the connection is closed."""
        if exc is None:  # pragma: no branch
            exc = ConnectionResetError('Connection closed')
        self.channel._set_exception(exc)
        self._connection_lost = exc
        self._task_process_out_msgs.cancel()

        # Wake up the writer if currently paused.
        if not self._paused:
            return
        waiter = self._drain_waiter
        if waiter is None:
            return
        self._drain_waiter = None
        if waiter.done():
            return
        waiter.set_exception(exc)

    def data_received(self, data):
        """Buffer incomming data until we have a complete message and then
        pass it to :class:`Channel`.

        Messages are fixed length.  The first four bytes (in network byte
        order) encode the length of the following payload.  The payload is
        a triple ``(msg_type, msg_id, content)`` encoded with the specified
        *codec*.

        """
        self._buffer.extend(data)
        while True:
            # We may have more then one message in the buffer,
            # so we loop over the buffer until we got all complete messages.

            if self._read_size is None and len(self._buffer) >= Header.size:
                # Received the complete header of a new message
                self._read_size = Header.unpack_from(self._buffer)[0]
                # TODO: Check for too large messages?
                self._read_size += Header.size

            if self._read_size and len(self._buffer) >= self._read_size:
                # At least one complete message is in the buffer
                data = self._buffer[Header.size:self._read_size]
                self._buffer = self._buffer[self._read_size:]
                self._read_size = None
                msg_type, msg_id, content = self.codec.decode(data)
                try:
                    self.channel._feed_data(msg_type, msg_id, content)
                except RuntimeError as exc:
                    self.channel._set_exception(exc)

            else:
                # No complete message in the buffer. We are done.
                break

    def eof_received(self):
        """Set a :exc:`ConnectionResetError` to the :class:`Channel`."""
        # In previous revisions, an IncompleteMessage error was raised if we
        # already received the beginning of a new message. However, having
        # to types of exceptions raised by this methods makes things more
        # complicated for the user. The benefit of the IncompleteMessage was
        # not big enough.
        self.channel._set_exception(ConnectionResetError())

    async def write(self, len_bytes, content):
        """Serialize *content* and write the result to the transport.

        This method is a `coroutine
        <https://docs.python.org/3/library/asyncio-task.html#coroutine>`_.

        """
        assert self._connection_lost is None
        content = len_bytes + content
        done = self._loop.create_future()
        self._out_msgs.put_nowait((done, content))
        await done

    def pause_writing(self):
        """Set the *paused* flag to ``True``.

        Can only be called if we are not already paused.

        """
        assert not self._paused
        self._paused = True
        if self._loop.get_debug():
            logger.debug("%r pauses writing", self)

    def resume_writing(self):
        """Set the *paused* flat to ``False`` and trigger the waiter future.

        Can only be called if we are paused.

        """
        assert self._paused
        self._paused = False
        if self._loop.get_debug():
            logger.debug("%r resumes writing", self)

        waiter = self._drain_waiter
        if waiter is not None:
            self._drain_waiter = None
            if not waiter.done():
                waiter.set_result(None)

    async def _process_out_msgs(self):
        try:
            while True:
                done, content = await self._out_msgs.get()
                self.transport.write(content)
                await self._drain_helper()
                done.set_result(None)
        except asyncio.CancelledError:
            assert self._connection_lost is not None

    async def _drain_helper(self, before=False):
        if self._connection_lost is not None:
            raise self._connection_lost
        if not self._paused:
            return
        waiter = self._drain_waiter
        assert waiter is None or waiter.cancelled()
        waiter = self._loop.create_future()
        self._drain_waiter = waiter
        await waiter


class Request:
    """Represents a request returned by :meth:`Channel.recv()`.  You shoudn't
    instantiate it yourself.

    *content* contains the incoming message.

    *msg_id* is the ID for that message.  It is unique within a channel.

    *protocol* is the channel's :class:`ChannelProtocol` instance that is used
    for writing back the reply.

    To reply to that request you can ``await`` :meth:`Request.reply()` or
    :meth:`Request.fail()`.

    """
    def __init__(self, content, message_id, protocol):
        self._content = content
        self._msg_id = message_id
        self._protocol = protocol

    @property
    def content(self):
        """The content of the incoming message."""
        return self._content

    async def reply(self, result):
        """Reply to the request with the provided result.

        This method is a `coroutine
        <https://docs.python.org/3/library/asyncio-task.html#coroutine>`_.

        """
        protocol = self._protocol
        if protocol._connection_lost is not None:
            raise protocol._connection_lost

        content = (RESULT, self._msg_id, result)
        try:
            content = protocol.codec.encode(content)
            msg_len = _get_checked_msg_len(content)

        except Exception as e:
            return await self.fail(e)
        else:
            await protocol.write(msg_len, content)

    async def fail(self, exception):
        """Indicate a failure described by the *exception* instance.

        This will raise a :exc:`~aiomas.exceptions.RemoteException` on the
        other side of the channel.

        This method is a `coroutine
        <https://docs.python.org/3/library/asyncio-task.html#coroutine>`_.

        """
        protocol = self._protocol
        if protocol._connection_lost is not None:
            raise protocol._connection_lost

        stacktrace = traceback.format_exception(exception.__class__, exception,
                                                exception.__traceback__)
        content = (EXCEPTION, self._msg_id, ''.join(stacktrace))
        content = protocol.codec.encode(content)
        msg_len = _get_checked_msg_len(content)
        await protocol.write(msg_len, content)


class Channel:
    """A Channel represents a request-reply channel between two endpoints. An
    instance of it is returned by :func:`open_connection()` or is passed to the
    callback of :func:`start_server()`.

    *protocol* is an instance of :class:`ChannelProtocol`.

    *transport* is an :class:`asyncio.BaseTransport`.

    *loop* is an instance of an :class:`asyncio.AbstractEventLoop`.

    """
    def __init__(self, protocol, codec, transport, loop):
        self._protocol = protocol
        self._codec = codec
        self._transport = transport
        self._loop = loop

        self._message_id = itertools.count()
        self._out_messages = {}  # message_id -> message
        self._in_queue = collections.deque()
        self._waiter = None  # A future.
        self._exception = None

    @property
    def codec(self):
        """The codec used to de-/encode messages send via the channel."""
        return self._codec

    @property
    def transport(self):
        """The transport of this channel (see the `Python documentation
        <https://docs.python.org/3/library/asyncio-protocol.html#transports>`_
        for details).

        """
        return self._transport

    def send(self, content):
        """Send a request *content* to the other end and return a future which
        is triggered when a reply arrives.

        One of the following exceptions may be raised:

        - :exc:`ValueError` if the message is too long (the length of the
          encoded message does not fit into a *long*, which is ~ 4 GiB).

        - :exc:`~aiomas.exceptions.RemoteException`: The remote site raised an
          exception during the computation of the result.

        - :exc:`ConnectionError` (or its subclass :exc:`ConnectionResetError`):
          The connection was closed during the request.

        - :exc:`RuntimeError`:

          - If an invalid message type was received.

          - If the future returned by this method was already triggered or
            canceled by a third party when an answer to the request arrives
            (e.g., if a task containing the future is cancelled).  You get
            more detailed exception messages if you `enable asyncio's debug
            mode`__

            __ https://docs.python.org/3/library/asyncio-dev.html

        .. code-block:: python

           try:
               result = await channel.request('ohai')
           except RemoteException as exc:
               print(exc)

        """
        if self._exception is not None:
            raise self._exception

        message_id = next(self._message_id)
        out_message = self._loop.create_future()
        if self._loop.get_debug():
            self._out_messages[message_id] = (out_message, content)
        else:
            self._out_messages[message_id] = out_message

        data = self._codec.encode((REQUEST, message_id, content))
        try:
            msg_len = _get_checked_msg_len(data)
        except ValueError:
            del self._out_messages[message_id]
            out_message.cancel()
            raise

        self._loop.create_task(self._protocol.write(msg_len, data))

        return out_message

    async def recv(self):
        """Wait for an incoming :class:`Request` and return it.

        May raise one of the following exceptions:


        - :exc:`ConnectionError` (or its subclass :exc:`ConnectionResetError`):
          The connection was closed during the request.

        - :exc:`RuntimeError`: If two processes try to read from the same
          channel or if an invalid message type was received.

        This method is a `coroutine
        <https://docs.python.org/3/library/asyncio-task.html#coroutine>`_.

        """
        if self._exception is not None:
            raise self._exception

        if not self._in_queue:
            if self._waiter is not None:
                raise RuntimeError('recv() called while another coroutine is '
                                   'already waiting for incoming data.')
            self._waiter = self._loop.create_future()
            try:
                await self._waiter
            finally:
                # In case of an exception, "self._waiter" is already set to
                # "None" by "self._set_exception()":
                self._waiter = None

        return self._in_queue.popleft()

    def _close(self):
        """Close the channel's transport."""
        if self._transport is not None:
            transport = self._transport
            self._transport = None
            return transport.close()

    async def close(self):
        """`Coroutine
        <https://docs.python.org/3/library/asyncio-task.html#coroutine>`_ that
        closes the channel and waits for all sub tasks to finish."""
        self._close()
        try:
            await self._protocol._task_process_out_msgs
            if self._loop.get_debug():
                futs = [m[0] for m in self._out_messages.values()]
            else:
                futs = self._out_messages.values()
            await asyncio.gather(*futs, return_exceptions=True)
        except asyncio.CancelledError:
            pass

    def get_extra_info(self, name, default=None):
        """Wrapper for :meth:`asyncio.BaseTransport.get_extra_info()`."""
        return self._transport.get_extra_info(name, default)

    def _feed_data(self, msg_type, msg_id, content):
        """Called by :class:`ChannelProtocol` when a new message arrived."""
        if msg_type == REQUEST:
            # Received new request
            message = Request(content, msg_id, self._protocol)
            self._in_queue.append(message)

            waiter = self._waiter
            if waiter is not None:
                self._waiter = None
                waiter.set_result(False)

        elif msg_type == RESULT or msg_type == EXCEPTION:
            # Received reply to a request
            if self._loop.get_debug():
                message, req = self._out_messages.pop(msg_id)
            else:
                message = self._out_messages.pop(msg_id)
            if message.done():
                errmsg = 'Request reply already set.'
                if message.cancelled():
                    errmsg = 'Request was cancelled.'
                if self._loop.get_debug():
                    errmsg += ' Request: %s; Reply: %s' % (req, content)
                raise RuntimeError(errmsg)

            if msg_type == RESULT:
                message.set_result(content)
            else:
                origin = self.get_extra_info('peername')
                message.set_exception(RemoteException(origin, content))

        else:
            raise RuntimeError('Invalid message type %d' % msg_type)

    def _set_exception(self, exc):
        """Set an exception as result for all futures managed by the Channel
        in order to stop all coroutines from reading/writing to the socket."""
        self._exception = exc

        # Set exception to wait-recv future
        waiter = self._waiter
        if waiter is not None:
            self._waiter = None
            if not waiter.cancelled():
                waiter.set_exception(exc)

        # Set exception to all message futures which wait for a reply
        for msg in self._out_messages.values():
            if self._loop.get_debug():
                msg, _ = msg
            if not msg.done():
                msg.set_exception(exc)

        self._close()



"""
This module contains some utility functions.

"""
import importlib
import ssl

import arrow


__all__ = [
    'arrow_serializer',
    'run',
    'make_ssl_client_context',
    'make_ssl_server_context',
    'obj_from_str',
]


def arrow_serializer():
    """Return a serializer for *arrow* dates.

    The return value is an argument tuple for
    :meth:`aiomas.codecs.Codec.add_serializer()`.

    """
    return arrow.Arrow, str, arrow.get


def run(until=None, loop=None):
    """Run the event loop forever or until the task/future *until* is finished.

    This is an alias to asyncio's ``run_forever()`` if *until* is ``None`` and
    to ``run_until_complete()`` if not.

    """
    if loop is None:
        import asyncio
        loop = asyncio.get_event_loop()
    if until is None:
        loop.run_forever()
    else:
        return loop.run_until_complete(until)


def make_ssl_server_context(cafile, certfile, keyfile):
    """Return an :class:`ssl.SSLContext` that can be used by a server socket.

    The server will use the certificate in *certfile* and private key in
    *keyfile* (both in PEM format) to authenticate itself.

    It requires clients to also authenticate themselves.  Their certificates
    will be validated with the root CA certificate in *cafile*.

    It will use *TLS 1.2* with *ECDH+AESGCM* encryption.  ECDH keys won't be
    reused in distinct SSL sessions.  Compression is disabled.

    """
    ctx = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
    ctx.set_ciphers('ECDH+AESGCM')
    ctx.load_cert_chain(certfile=certfile, keyfile=keyfile)
    ctx.verify_mode = ssl.CERT_REQUIRED
    ctx.load_verify_locations(cafile=cafile)
    ctx.options |= ssl.OP_SINGLE_ECDH_USE
    ctx.options |= ssl.OP_NO_COMPRESSION
    return ctx


def make_ssl_client_context(cafile, certfile, keyfile):
    """Return an :class:`ssl.SSLContext` that can be used by a client socket.

    It uses the root CA certificate in *cafile* to validate the server's
    certificate.  It will also check the server's hostname.

    The client will use the certificate in *certfile* and private key in
    *keyfile* (both in PEM format) to authenticate itself.

    It will use *TLS 1.2* with *ECDH+AESGCM* encryption.

    """
    ctx = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
    ctx.set_ciphers('ECDH+AESGCM')
    ctx.verify_mode = ssl.CERT_REQUIRED
    ctx.load_verify_locations(cafile=cafile)
    ctx.check_hostname = True
    ctx.load_cert_chain(certfile=certfile, keyfile=keyfile)
    return ctx


def obj_from_str(obj_path):
    """Return the object that the string *obj_path* points to.

    The format of *obj_path* is ``mod:obj`` where *mod* is a (possibly nested)
    module name and *obj* is an ``.`` separate object path, for example::

        module:Class
        module:Class.function
        package.module:Class
        package.module:Class.function

    Raise a :exc:`ValueError` if the *obj_path* is malformed, an
    :exc:`ImportError` if the module cannot be imported or an
    :exc:`AttributeError` if an object does not exist.

    """
    try:
        mod_name, obj_names = obj_path.split(':')
    except ValueError:
        raise ValueError('Malformed object name "{}": Expected "module:object"'
                         .format(obj_path)) from None

    obj_names = obj_names.split('.')
    obj = importlib.import_module(mod_name)
    for name in obj_names:
        obj = getattr(obj, name)

    return obj



"""
This module helps you to :func:`start()` an agent container in a new
subprocess.  The new container will have a :class:`Manager` agent that allows
the master process to spawn other agents in the new container.

The following example demonstrate how you can build a nice CLI with the `click
<http://click.pocoo.org>`_ around this module. The script will start you
a container with an :class:`~aiomas.clocks.ExternalClock` and the
:class:`~aiomas.codecs.MsgPackBlosc` codec:

.. code-block:: python3

   # container.py
   import logging

   import aiomas
   import arrow
   import click


   def validate_addr(ctx, param, value):
       try:
           host, port = value.rsplit(':', 1)
           return (host, int(port))
       except ValueError as e:
           raise click.BadParameter(e)


   def validate_start_date(ctx, param, value):
       try:
           arrow.get(value)  # Check if the date can be parsed
       except arrow.parser.ParserError as e:
           raise click.BadParameter(e)
       return value

   @click.command()
   @click.option('--start-date', required=True,
                 callback=validate_start_date,
                 help='Start date for the agents (ISO-8601 compliant, e.g.: '
                      '2010-03-27T00:00:00+01:00')
   @click.option('--log-level', '-l', default='info', show_default=True,
                 type=click.Choice(['debug', 'info', 'warning', 'error',
                                    'critical']),
                 help='Log level for the MAS')
   @click.argument('addr', metavar='HOST:PORT', callback=validate_addr)
   def main(addr, start_date, log_level):
       logging.basicConfig(level=getattr(logging, log_level.upper()))
       clock = aiomas.ExternalClock(start_date, init_time=-1)
       codec = aiomas.codecs.MsgPackBlosc
       task = aiomas.subproc.start(addr, clock=clock, codec=codec)
       aiomas.run(until=task)


   if __name__ == '__main__':
       main()

Example usage: :command:`python container.py
--start-date=2010-03-27T00:00:00+01:00 localhost:5556`.

.. note::

   You should use ``sys.executable`` instead of just ``'python'`` when you
   start a new subprocess from within a Python script to make sure you use the
   correct (same) interpreter.

"""
import asyncio
import logging

from . import agent, rpc, util


logger = logging.getLogger(__name__)


async def start(addr, **container_kwargs):
    """`Coroutine
    <https://docs.python.org/3/library/asyncio-task.html#coroutine>`_ that
    starts a container with a :class:`Manager` agent.

    The agent will connect to *addr* ``('host', port)`` and wait for commands
    to spawn new agents within its container.

    The *container_kwargs* will be passed to
    :meth:`aiomas.agent.Container.create()` factory function.

    This coroutine finishes after :meth:`Manager.stop()` was called or when
    a :exc:`KeyboardInterrupt` is raised.

    """
    container_kwargs.update(as_coro=True)
    container = await agent.Container.create(addr, **container_kwargs)
    try:
        manager = Manager(container)
        await manager.stop_received
    except KeyboardInterrupt:
        logger.info('Execution interrupted by user')
    finally:
        await container.shutdown(as_coro=True)


class Manager(agent.Agent):
    """An agent that can start other agents within its container.

    If the container uses an :class:`~aiomas.clocks.ExternalClock`, it can also
    set the time for the container's clock.

    """
    def __init__(self, container):
        super().__init__(container)
        self.stop_received = container.loop.create_future()

    @rpc.expose
    async def spawn(self, qualname, *args, **kwargs):
        """Create a new instance of an agent and return a proxy to it and its
        address.

        *qualname* is a string defining a class (or factory method/coroutine)
        for instantiating the agent (see :func:`aiomas.util.obj_from_str()` for
        details).  *args* and *kwargs* get passed to this callable as
        positional and keyword arguemnts respectively.

        This is an exposed `coroutine
        <https://docs.python.org/3/library/asyncio-task.html#coroutine>`_.

        """

        callable = util.obj_from_str(qualname)
        if asyncio.iscoroutinefunction(callable):
            agent = await callable(self.container, *args, **kwargs)
        else:
            agent = callable(self.container, *args, **kwargs)

        logger.debug('Spawned %s(%s)', agent.__class__.__name__, agent)

        return agent, agent.addr

    @rpc.expose
    def set_time(self, time):
        """Set the agent's container's time to *time*.

        This only works if the container uses an
        :class:`~aiomas.clocks.ExternalClock`.

        This is an exposed function.

        """
        self.container.clock.set_time(time)

    @rpc.expose
    def stop(self):
        """Triggers the *stop_received* future of this agent causing its
        container process to shutodwn and terminate.

        This is an exposed function.

        """
        self.stop_received.set_result(True)



"""
This module implements the base class for agents (:class:`Agent`) and
containers for agents (:class:`Container`).

Every agent must live in a container.  A container can contain one ore more
agents.  Containers are responsible for making connections to other containers
and agents.  They also provide a factory function for spawning new agent
instances and registering them with the container.

Thus, the :class:`Agent` base class is very light-weight.  It only has a name,
a reference to its container and an RPC router (see :mod:`aiomas.rpc`).

"""
import asyncio
import collections
import weakref
import socket
import ssl as sslmod

from . import channel, clocks, local_queue, rpc, util


__all__ = ['SSLCerts', 'Container', 'Agent']


PROTOCOLS = {
    'tcp',  # TCP sockets
    'ipc',  # Inter Process Communication with Unix domain sockets
    'local',  # LocalQueue transport
}


SSLCerts = collections.namedtuple('SSLCerts', 'cafile, certfile, keyfile')
""":func:`~collections.namedtuple` storing the names of a CA file, a
certificate file and the associated private key file.

See also :func:`aiomas.util.make_ssl_server_context()` and
:func:`aiomas.util.make_ssl_client_context()`.

"""


def _get_base_url(addr):
    # Get base URL for agents (tcp or ipc)
    if isinstance(addr, tuple):
        host, port = addr
        if host in [None, '', '::', '0.0.0.0']:
            host = socket.getfqdn()
        base_url = '%s://%s:%s/' % ('tcp', host, port)
    elif isinstance(addr, local_queue.LocalQueue):
        base_url = 'local://%s/' % addr.queue_id
    else:
        base_url = '%s://[%s]/' % ('ipc', addr)

    return base_url


def _make_ssl_contexts(ssl):
    """Derive and return a tuple *(Server SSLContext, Client SSLContext)* from
    *ssl*.

    *ssl* may either be a :class:`SSLCerts` instance or a tuple of two
    :class:`ssl.SSLContext` instances.

    In other cases, return ``(None, None)``.

    """
    if isinstance(ssl, SSLCerts):
        ssl_server_ctx = util.make_ssl_server_context(**ssl._asdict())
        ssl_client_ctx = util.make_ssl_client_context(**ssl._asdict())
    elif isinstance(ssl, tuple):
        if (len(ssl) != 2 or
                not isinstance(ssl[0], sslmod.SSLContext) or
                not isinstance(ssl[1], sslmod.SSLContext)):
            raise TypeError('"ssl" must contain two "ssl.SSLContext" '
                            'instances; one for the server and one for '
                            'the client.')
        ssl_server_ctx = ssl[0]
        ssl_client_ctx = ssl[1]
    else:
        ssl_server_ctx = None
        ssl_client_ctx = None

    return ssl_server_ctx, ssl_client_ctx


class Container:
    """Container for agents.

    You should not instantiate containers directly but use the :meth:`create()`
    method/coroutine instead.  This makes sure that the container's server
    socket is fully operational when it is created.

    The container allows its agents to create connections to other agents (via
    :meth:`connect()`).

    In order to destroy a container and close all of its sockets, call
    :meth:`shutdown()`.

    """
    router = rpc.Service(['agents'])

    @classmethod
    def create(cls, addr, *, clock=None, codec=None, extra_serializers=None,
               ssl=None, as_coro=False, loop=None):
        """Instantiate a container and create a server socket for it.

        This function is a classmethod and `coroutine
        <https://docs.python.org/3/library/asyncio-task.html#coroutine>`_.

        :param addr:
            is the address that the server socket is bound to.  It may be
            a ``(host, port)`` tuple for a TCP socket, a path for a Unix domain
            socket, or a *LocalQueue* instance as returned by the
            :func:`aiomas.local_queue.get_queue()` function.

            **TCP sockets**

            If ``host`` is ``'0.0.0.0'`` or ``'::'``, the server is bound to
            all available IPv4 *or* IPv6 interfaces respectively.  If ``host``
            is ``None`` or ``''``, the server is bound to all available IPv4
            *and* IPv6 interfaces.  In these cases, the machine's FQDN (see
            :func:`socket.getfqdn()`) should be resolvable and point to that
            machine as it will be used for the agent's addresses.

            If ``host`` is a simple (IPv4 or IPv6) IP address, it will be used
            for the agent's addresses as is.

            **LocalQueue**

            In contrast to TCP, multiple LocalQueue connections between
            containers (within the same thread and OS process) send and receive
            message in a deterministic order, which is useful for testing and
            debugging.

            LocalQueue instances should be retrieved via the
            :func:`aiomas.local_queue.get_queue()` function (which also
            available as ``aiomas.get_queue()``). This function always
            returns the same instance for a given queue ID.

        :param clock:
            can be an instance of :class:`~aiomas.clocks.BaseClock`.

            It allows you to decouple the container's (and thus, its agent's)
            time from the system clock.  This makes it easier to integrate your
            system with other simulators that may provide a clock for you or to
            let your MAS run as fast as possible.

            By default, the real-time :class:`~aiomas.clocks.AsyncioClock` will
            be used.

        :param codec:
            can be a :class:`~aiomas.codecs.Codec` subclass (not an instance!).
            :class:`~aiomas.codecs.JSON` is used by default.

        :param extra_serializers:
            is an optional list of extra serializers for the codec.  The list
            entries need to be callables that return a tuple with the arguments
            for :meth:`~aiomas.codecs.Codec.add_serializer()`.

        :param ssl:
            allows you to enable TLS for all incoming and outgoing TCP
            connections.  It may either be an :class:`SSLCerts` instance or
            a tuple containing two :class:`~ssl.SSLContext` instances, where
            the first one will be used for the server socket, the second one
            for client sockets.

        :param as_coro:
            must be set to ``True`` if the event loop is already running when
            you call this method.  This function then returns a coroutine that
            you need to ``await`` in order to get the container.  By default it
            will block until the server has been started and return the
            container.

        :param loop:
            The asyncio event loop to use.  If ``None``, the default one will
            be used.

        :return:
            a fully initialized :class:`Container` instance if *async* is
            ``False`` or else a coroutine returning the instance when it is
            done.

        Invocation examples::

            # Synchronous:
            container = Container.create(...)

            # Asynchronous:
            container = await Container.create(..., as_coro=True)


        """
        base_url = _get_base_url(addr)
        ssl_server_ctx, ssl_client_ctx = _make_ssl_contexts(ssl)

        # Get default codec and clock if none were provided
        if loop is None:
            loop = asyncio.get_event_loop()
        if codec is None:
            codec = channel.DEFAULT_CODEC
        if clock is None:
            clock = clocks.AsyncioClock(loop=loop)

        # Prepend the Arrow date serializer to the list of serializers
        if extra_serializers is None:
            extra_serializers = []
        extra_serializers = [util.arrow_serializer] + extra_serializers

        # Additional keyword arguments for connecting to other containers
        connect_kwargs = {
            'codec': codec,
            'extra_serializers': extra_serializers,
            'ssl': ssl_client_ctx,
        }

        # Actually instantiate the container and start the server socket
        async def _start():
            container = cls(base_url, loop, clock, connect_kwargs)
            tcp_server = await rpc.start_server(
                addr,
                container,
                client_connected_cb=container._add_to_con_cache,
                codec=codec,
                extra_serializers=extra_serializers,
                ssl=ssl_server_ctx,
                loop=loop,
            )
            container.set_server(tcp_server)
            return container

        if as_coro:
            return _start()
        else:
            return util.run(_start(), loop=loop)

    def __init__(self, base_url, loop, clock, connect_kwargs):
        self._tcp_server = None
        self._rpc_cons = set()
        self._loop = loop
        self._base_url = base_url
        self._clock = clock
        self._connect_kwargs = connect_kwargs

        # The agents managed by this container.
        # The agents' routers are subrouters of the container's root router.
        self.agents = rpc.ServiceDict()

        # Caches
        self._connections_out_futs = {}  # Futures for outgoing connections
        self._connections_out = {}  # RPC connections to containers by address
        self._remote_agent_futs = {}  # Futures for remote agent validation
        self._remote_agents = {}  # Validated remote agents by connection

    def __str__(self):
        return '%s(%r)' % (self.__class__.__name__, self._base_url)

    @property
    def loop(self):
        """The event loop the container is using."""
        return self._loop

    @property
    def clock(self):
        """The clock of the container.  Instance of
        :class:`aiomas.clocks.BaseClock`."""
        return self._clock

    def set_server(self, server):
        if self._tcp_server is not None:
            raise RuntimeError('Server already set.')
        self._tcp_server = server

    async def connect(self, url, timeout=0):
        """Connect to the argent available at *url* and return a proxy to it.

        *url* is a string ``<protocol>://<addr>//<agent-id>`` (e.g.,
        ``'tcp://localhost:5555/0'``).

        With a *timeout* of 0 (the default), there will only be one connection
        attempt before an error is raised (:exc:`ConnectionRefusedError` for
        TCP sockets and LocalQueue, :exc:`FileNotFoundError` for Unix domain
        sockets).  If you set *timeout* to a number > 0 or ``None``, this
        function will try to connect repeatedly for at most that many seconds
        (or indefinitely) before an error is raised.  Use this if the remote
        agent's container may not yet exist.

        This function is a `coroutine
        <https://docs.python.org/3/library/asyncio-task.html#coroutine>`_.

        """
        addr, aid = self._parse_url(url)

        rpc_con = await self._open_connection(addr, timeout)
        remote_agent = await self._validate_aid(aid, rpc_con, addr, url)

        return remote_agent

    def shutdown(self, as_coro=False):
        """Close the container's server socket and the RPC services for all
        outgoing TCP connections.

        If *async* is left to ``False``, this method calls
        :meth:`asyncio.AbstractEventLoop.run_until_complete()` in order to wait
        until all sockets are closed.

        Set *async* to ``True`` if the event loop is already running (e.g.,
        because you are in a coroutine).  The return value then is a `coroutine
        <https://docs.python.org/3/library/asyncio-task.html#coroutine>`_ that
        you need to ``await`` in order to actually shut the container down::

            await container.shutdown(as_coro=True)

        """
        async def _shutdown():
            # Close all outgoing connections
            futs = []
            for con in self._connections_out.values():
                futs.append(con.close())
            await asyncio.gather(*futs, loop=self._loop)

            if self._tcp_server:
                # Request closing the server socket and cancel the services
                self._tcp_server.close()
                for con in self._rpc_cons:
                    con.service.cancel()

                # Wait for server and services to actually terminate
                await asyncio.gather(
                    self._tcp_server.wait_closed(),
                    *[c.service for c in self._rpc_cons],
                    return_exceptions=True,
                    loop=self._loop,
                )

                self._tcp_server = None
                self._rpc_cons = None

        if as_coro:
            return _shutdown()
        else:
            util.run(_shutdown(), loop=self._loop)

    @router.expose
    def validate_aid(self, aid):
        """Return the class name for the agent represented by *aid* if it
        exists or ``None``."""
        agents = self.agents.dict
        if aid in agents:
            return agents[aid].__class__.__name__

    def _add_to_con_cache(self, rpc_con):
        """Client-connected-callback for the server socket that adds the RPC
        connection *rpc_con* to the set of connections."""
        self._rpc_cons.add(rpc_con)

    @staticmethod
    def _parse_url(url):
        """Parse the agent *url* and return a ``((host, port), agent)`` tuple.

        Raise a :exc:`ValueError` if the URL cannot be parsed.

        """
        try:
            proto, addr_aid = url.split('://', 1)
            assert proto in PROTOCOLS, '{} not in {}'.format(proto, PROTOCOLS)

            if proto == 'tcp':
                addr, aid = addr_aid.split('/', 1)
                host, port = addr.rsplit(':', 1)
                if host[0] == '[' and host[-1] == ']':
                    # IPv6 addresses may be surrounded by []
                    host = host[1:-1]
                addr = (host, int(port))

            elif proto == 'ipc':
                assert addr_aid[0] == '['
                addr, aid = addr_aid[1:].split(']/', 1)

            elif proto == 'local':
                queue_id, aid = addr_aid.split('/', 1)
                addr = local_queue.get_queue(queue_id)
                aid = aid

            assert aid, 'No agent ID specified.'

        except (AssertionError, IndexError, ValueError) as e:
            raise ValueError('Cannot parse agent URL "{}": {}'.format(url, e))

        return addr, aid

    async def _open_connection(self, addr, timeout):
        if addr in self._connections_out:
            # Return cached connection
            rpc_con = self._connections_out[addr]
        elif addr in self._connections_out_futs:
            # Wait for ongoing connection attempt
            rpc_con = await self._connections_out_futs[addr]
        else:
            # Open new connection
            fut = self._loop.create_future()
            self._connections_out_futs[addr] = fut

            rpc_con = await rpc.open_connection(
                addr,
                rpc_service=self,
                timeout=timeout,
                loop=self._loop,
                **self._connect_kwargs,
            )

            # Put connection into the cache
            self._rpc_cons.add(rpc_con)
            self._connections_out[addr] = rpc_con

            # Trigger future and remove it from the cache
            fut.set_result(rpc_con)
            self._connections_out_futs.pop(addr)

            # Initialize caches for remote agents
            self._remote_agents[rpc_con] = weakref.WeakValueDictionary()
            self._remote_agent_futs[rpc_con] = {}

        return rpc_con

    async def _validate_aid(self, aid, rpc_con, addr, url):
        remote_agents = self._remote_agents[rpc_con]
        remote_agent_futs = self._remote_agent_futs[rpc_con]

        if aid in remote_agents:
            remote_agent = remote_agents[aid]
        elif aid in remote_agent_futs:
            remote_agent = await remote_agent_futs[aid]
        else:
            fut = self._loop.create_future()
            remote_agent_futs[aid] = fut

            remote_type = await rpc_con.remote.validate_aid(aid)
            if remote_type is None:
                raise ConnectionError('Agent "{}" does not exist in '
                                      'Container({!r})'.format(aid, addr))
            remote_agent = getattr(rpc_con.remote.agents, aid)
            remote_agent._str = '{}Proxy({!r})'.format(remote_type, url)

            remote_agents[aid] = remote_agent
            fut.set_result(remote_agent)
            remote_agent_futs.pop(aid)

        return remote_agent

    def _register(self, agent):
        """Register *agent* with the container."""
        aid = str(len(self.agents.dict))
        url = self._base_url + aid
        self.agents.dict[aid] = agent
        self.agents.router.set_sub_router(agent.router, aid)
        return url


class Agent:
    """Base class for all agents."""

    router = rpc.Service()
    """Descriptor that creates an RPC :class:`~aiomas.rpc.Router` for every
    agent instance.

    You can override this in a sub-class if you need to.  (Usually, you don't.)

    """
    def __init__(self, container):
        if not isinstance(container, Container):
            raise TypeError('"container" must be a "Container" instance but '
                            'is {}'.format(container))
        addr = container._register(self)
        self.__container = container
        self.__addr = addr
        self.__name = '{}({!r})'.format(self.__class__.__name__, addr)

    def __str__(self):
        return self.__name

    @property
    def container(self):
        """The :class:`Container` that the agent lives in."""
        return self.__container

    @property
    def addr(self):
        """The agent's address."""
        return self.__addr



"""
Clocks to be used with :class:`aiomas.agent.Container`.

All clocks should subclass :class:`BaseClock`.  Currently available clock types
are:

- :class:`AsyncioClock`: a real-time clock synchronized with the :mod:`asyncio`
  event loop.

- :class:`ExternalClock`: a clock that can be set by external tasks / processes
  in order to synchronize it with external systems or simulators.

"""
import asyncio
import heapq
import itertools

import arrow


__all__ = ['BaseClock', 'AsyncioClock', 'ExternalClock', 'TimerHandle']


class BaseClock:
    """Interface for clocks.

    Clocks must at least implement :meth:`time()` and :meth:`utcnow`.

    """

    def time(self):
        """Return the value (in seconds) of a monotonic clock.

        The return value of consecutive calls is guaranteed to be greater or
        equal then the results of previous calls.

        The initial value may not be defined. Don't depend on it.

        """
        raise NotImplementedError()

    def utcnow(self):
        """Return an :class:`arrow.arrow.Arrow` date with the current time in
        UTC."""
        raise NotImplementedError()

    def sleep(self, dt, result=None):
        """Sleep for a period *dt* in seconds.  Return an
        :class:`asyncio.Future`.

        If *result* is provided, it will be passed back to the caller when
        the coroutine has finished.

        """
        raise NotImplementedError()

    def sleep_until(self, t, result=None):
        """Sleep until the time *t*.  Return an :class:`asyncio.Future`.

        *t* may either be a number in seconds or an :class:`arrow.arrow.Arrow`
        date.

        If *result* is provided, it will be passed back to the caller when
        the coroutine has finished.

        """
        raise NotImplementedError()

    def call_in(self, dt, func, *args):
        """Schedule the execution of ``func(*args)`` in *dt* seconds and return
        immediately.

        Return an opaque handle which lets you cancel the scheduled call via
        its ``cancel()`` method.

        """
        raise NotImplementedError()

    def call_at(self, t, func, *args):
        """Schedule the execution of ``func(*args)`` at *t* and return
        immediately.

        *t* may either be a number in seconds or an :class:`arrow.arrow.Arrow`
        date.

        Return an opaque handle which lets you cancel the scheduled call via
        its ``cancel()`` method.

        """
        raise NotImplementedError()

    def _check_date(self, date):
        """Assert that *date* is not in the past and convert it into float if
        it is an :class:`arrow.arrow.Arrow`."""
        if isinstance(date, arrow.arrow.Arrow):
            t = (date - self.utcnow()).total_seconds() + self.time()
        else:
            t = date
        if t <= self.time():
            raise ValueError('Date "{}" is in the past'.format(date))
        return t


class AsyncioClock(BaseClock):
    """:mod:`asyncio` based real-time clock."""
    def __init__(self, loop=None):
        if loop is None:
            loop = asyncio.get_event_loop()
        self._loop = loop

    def time(self):
        return self._loop.time()

    def utcnow(self):
        return arrow.utcnow()

    def sleep(self, dt, result=None):
        return asyncio.sleep(dt, result, loop=self._loop)

    def sleep_until(self, t, result=None):
        t = self._check_date(t)
        return asyncio.sleep(t - self.time(), result, loop=self._loop)

    def call_in(self, dt, func, *args):
        return self._loop.call_later(dt, func, *args)

    def call_at(self, t, func, *args):
        t = self._check_date(t)
        return self._loop.call_at(t, func, *args)


class ExternalClock(BaseClock):
    """A clock that can be set by external process in order to synchronize
    it with other systems.

    The initial UTC date *utc_start* may either be an
    :class:`arrow.arrow.Arrow` instance or something that
    :meth:`arrow.factory.ArrowFactory.get()` can parse.

    """
    def __init__(self, utc_start, init_time=0, loop=None):
        if not isinstance(utc_start, arrow.arrow.Arrow):
            utc_start = arrow.get(utc_start).to(tz='UTC')

        if loop is None:
            loop = asyncio.get_event_loop()

        self._time = init_time
        self._utc_start = utc_start
        self._loop = loop

        self._queue = []
        self._eid = itertools.count()

    def time(self):
        return self._time

    def utcnow(self):
        return self._utc_start.replace(seconds=self._time)

    def set_time(self, t):
        if t <= self._time:
            raise ValueError('Time must be > {} but is {}'
                             .format(self._time, t))
        self._time = t

        while self._queue and self._queue[0][0] <= t:
            _, _, future, result = heapq.heappop(self._queue)
            if not future.cancelled():
                future.set_result(result)

    def sleep(self, dt, result=None):
        if dt <= 0:
            raise ValueError('dt must be > 0 but is {}'.format(dt))
        return self.sleep_until(self._time + dt, result)

    def sleep_until(self, t, result=None):
        t = self._check_date(t)
        f = self._loop.create_future()
        heapq.heappush(self._queue, (t, next(self._eid), f, result))
        return f

    def call_in(self, dt, func, *args):
        if dt <= 0:
            raise ValueError('dt must be > 0 but is {}'.format(dt))
        return self.call_at(self._time + dt, func, *args)

    def call_at(self, t, func, *args):
        def cb(fut):
            func(*args)

        t = self._check_date(t)
        f = self.sleep_until(t)
        f.add_done_callback(cb)
        return TimerHandle(f, cb)


class TimerHandle:
    """This class lets you cancel calls scheduled by :class:`ExternalClock`."""
    def __init__(self, future, callback):
        self._future = future
        self._callback = callback

    def cancel(self):
        """Cancel the scheduled call represented by this handle."""
        self._future.remove_done_callback(self._callback)
        self._future.cancel()



import pytest
import aiomas

#
# Production code (exampleagent.py)
#


class ExampleAgent(aiomas.Agent):
    async def run(self, target_addr, num):
        remote_agent = await self.container.connect(target_addr)
        return (await remote_agent.service(num))

    @aiomas.expose
    async def service(self, val):
        await self.container.clock.sleep(0.001)
        return val


#
# Testing code (test_exampleagent.py)
#


@pytest.yield_fixture
def container(event_loop, unused_tcp_port):
    """This fixture creates a new Container instance for every test and binds
    it to a random port.

    It requires the *event_loop" fixture, so every test will also have a fresh
    event loop.

    """
    # Create container and bind its server socket to a random port:
    c = aiomas.Container.create(('127.0.0.1', unused_tcp_port))

    # Yield the container to the test case:
    yield c

    # Clean-up that is run after the test finished:
    c.shutdown()

    # The "event_loop" fixture closes the event loop after the test, but we
    # need an existing event loop for the following tests, so start a new one:
    import asyncio
    asyncio.set_event_loop(asyncio.new_event_loop())


# The "@pytest.mark.asyncio" decorator allows you do use "await"/"yield from"
# directly within your test case.
#
# The "container" argument tells pytest to pass the return/yield value of the
# corresponding fixture to your test.
@pytest.mark.asyncio
async def test_example_agent(container):
    num = 42
    # Start two agents:
    agents = [ExampleAgent(container) for _ in range(2)]
    # Run the 1st one and let it connect to the 2nd one.  Check the return
    # value of the 1st one's run() task:
    res = await agents[0].run(agents[1].addr, num)
    assert res == num



"""
This module implements remote procedure calls (RPC) on top of request-reply
channels (see :mod:`aiomas.channel`).

RPC connections are represented by instances of :class:`RpcClient` (one for
each side of a :class:`aiomas.channel.Channel`).  They provide access to the
functions served by the peer via :class:`Proxy` instances.  Optionally, they
can provide their own RPC service so that the peer can make calls as well.

An RPC service is an object with a ``router`` attribute which is an instance of
:class:`Router`.  A router resolves paths requested by the peer.  It can also
handle sub-routers (which allows you to build hierarchies for nested calls) and
is able to perform a reverse-lookup of a router (mapping a fuction to its
path).

Routers an be attached to both, classes and dictionaries with functions.
Dictionaires need to be wrapped with a :class:`ServiceDict`.  Classes need to
have a :class:`Service` class attribute named ``router``.  :class:`Service` is
a descriptor which creates a :class:`Router` for every instance of that class.

Functions that should be callable from the remote side must be decorated with
:func:`expose()`; :func:`Router.expose()` and :func:`Service.expose()` are
aliases for it.

"""
import asyncio
import logging
import weakref

from . import channel, exceptions


__all__ = [
    'open_connection', 'start_server', 'rpc_service_process', 'expose',
    '_handle_request',
    'Service', 'ServiceDict', 'Router', 'RpcClient', 'Proxy',
]


logger = logging.getLogger(__name__)


async def open_connection(addr, *, rpc_service=None, **kwds):
    """Return an :class:`RpcClient` connected to *addr*.

    This is a convenience wrapper for :meth:`aiomas.channel.open_connection()`.
    All keyword arguments *(kwds)* are forwared to it.

    You can optionally pass a *rpc_service* to allow the peer to call back to
    us.

    This function is a `coroutine
    <https://docs.python.org/3/library/asyncio-task.html#coroutine>`_.

    """
    if rpc_service is not None and not _is_rpc_service(rpc_service):
        raise ValueError('"{}" is not a valid RPC service'.format(rpc_service))

    c = await channel.open_connection(addr, **kwds)
    return RpcClient(c, rpc_service)


async def start_server(addr, rpc_service, client_connected_cb=None, **kwds):
    """Start a server socket on *host:port* and create an RPC service with
    the provided *handler* for each new client.

    This is a convenience wrapper for :meth:`aiomas.channel.start_server()`.
    All keyword arguments *(kwds)* are forwared to it.

    *rpc_service* must be an RPC service (an object with a ``router`` attribute
    that is an instance of :class:`Router`).

    *client_connected_cb* is an optional callback that will be called with
    with the :class:`RpcClient` instance for each new connection.

    Raise a :exc:`ValueError` if *handler* is not decorated properly.

    This function is a `coroutine
    <https://docs.python.org/3/library/asyncio-task.html#coroutine>`_.

    """
    if not _is_rpc_service(rpc_service):
        raise ValueError('"{}" is not a valid RPC service.'
                         .format(rpc_service))

    def fac(channel):
        """Create an RPC client for each new connection and call the
        *client_connected_cb* if there is one."""
        rpc_cli = RpcClient(channel, rpc_service)
        if client_connected_cb:
            client_connected_cb(rpc_cli)
        return rpc_cli

    server = await channel.start_server(addr, fac, **kwds)
    return server


async def rpc_service_process(rpc_client, router, channel):
    """RPC service process for a connection *rpc_lient*.

    Serves the functions provided by the :class:`Router` *router* via the
    :class:`~aiomas.channel.Channel` *channel*.

    Forward errors raised by the handler to the caller.

    Stop running when the connection closes.

    This function is a `coroutine
    <https://docs.python.org/3/library/asyncio-task.html#coroutine>`_.

    """
    loop = channel._loop

    try:
        while True:
            # Wait for a request
            request = await channel.recv()

            # Dispatch handling of the request to a sub task to make dead-locks
            # less likely when multiple agents share the same connection.
            loop.create_task(_handle_request(request, loop, router))

    except asyncio.CancelledError:
        pass

    except ConnectionError as ce:
        rpc_client._connection_reset(ce)

    finally:
        try:
            await channel.close()
        except RuntimeError:
            # May happen when the loop is already closed
            pass


async def _handle_request(request, loop, router):
    """Handle the *request*.

    Resolve the path, execute the corresponding message and set the result
    or exception.

    """
    path, args, kwargs = request.content
    # logger.debug('Request: [%s, %s, %s]' % (path, args, kwargs))

    # Resolve requested path
    try:
        func = router.resolve(path)
    except LookupError as exc:
        await request.fail(exc)
        return

    # Call requested function
    try:
        res = func(*args, **kwargs)
        if asyncio.iscoroutine(res):
            res = await loop.create_task(res)
        reply = request.reply
    except Exception as e:
        reply = request.fail
        res = e

    try:
        await reply(res)
    except ConnectionResetError:
        pass


def _is_rpc_service(obj):
    if not hasattr(obj, 'router'):
        return False

    if not isinstance(obj.router, Router):
        return False

    return True


def expose(func):
    """Decorator that enables RPC access to the decorated function.

    *func* will not be wrapped but only gain an ``__rpc__`` attribute.

    """
    if not hasattr(func, '__call__'):
        raise ValueError('"{}" is not callable.'.format(func))

    func.__rpc__ = True
    return func


class ServiceDict:
    """Wrapper for dicts so that they can be used as RPC routers."""
    __rpc__ = True

    def __init__(self, dict=None):
        self.dict = {} if dict is None else dict
        """The wrapped dict."""

        self.router = Router(self)
        """The dict's router instance."""

        self.__getrpc__ = self.dict.__getitem__
        for key, val in self.dict.items():
            # Iterate over all entries and look for objects with routers.
            # Set *router* as parent to these sub-routers.
            if hasattr(val, 'router'):
                Router.set_sub_router(self.router, val.router, key)


class Service:
    """A Data Descriptor that creates a new :class:`Router` instance for each
    class instance to which it is set.

    The attribute name for the Service should always be *router*::

        class Spam:
            router = aiomas.rpc.Service()

    You can optionally pass a list with the attribute names of classes with
    sub-routers.  This required to build hierarchies of routers, e.g.::

        class Eggs:
            router = aiomas.rpc.Service()


        class Spam:
            router = aiomas.rpc.Service(['eggs'])

            def __init__(self):
                self.eggs = Eggs()  # Instance with a sub-router

    """
    def __init__(self, sub_routers=()):
        self._sub_router_names = sub_routers

    def __set__(self, instance, value):
        """Raise :exc:`AttributeError` to forbid overwriting this attribute."""
        raise AttributeError('Read-only attribute.')

    def __get__(self, instance, cls):
        """If accessed from the class, return this Service instance.  If
        accessed from an *instance*, return the :class:`Router` instance for
        *instance*.

        """
        if instance is None:
            return self

        if 'router' not in instance.__dict__:
            # Create new Router for "instance" and add all sub-router:
            router = instance.__dict__.setdefault('router', Router(instance))
            for name in self._sub_router_names:
                router.add(name)

        return instance.__dict__['router']

    expose = staticmethod(expose)
    """Alias for :func:`expose()`."""


class Router:
    """The Router resolves paths to functions provided by their object *obj*
    (or its children).  It can also perform a reverse lookup to get the path
    of the router (and the router's *obj*).

    The *obj* can be a class, an instance or a dict.

    """
    def __init__(self, obj):
        # Mark *obj* as node in the RPC hierarchy and and create an alias
        # for accessing its child elements.
        obj.__rpc__ = True
        obj.__getrpc__ = obj.__getattribute__

        self.obj = obj  #: The object to which this router belongs to.
        self.name = ''  #: The name of the router (empty for root routers).
        self.parent = None  #: The parent router or ``None`` for root routers.

        self._cache = {}  # Maps already resolved paths to functions

    @property
    def path(self):
        """The path to this router (without trailing slash)."""
        router = self
        parts = []
        while router.parent is not None:
            # We go from child to root here
            parts.append(router.name)
            router = router.parent

        return '/'.join(reversed(parts))  # Reverse to get root first

    def resolve(self, path):
        """Resolve *path* and return the corresponding function.

        *path* is a string with path components separated by */* (without
        trailing slash).

        Raise a :exc:`LookupError` if no handler function can be found for
        *path* or if the function is not exposed (see :func:`expose()`).

        """
        try:
            obj = self._cache[path]
        except KeyError:
            parent = None
            obj = self.obj
            parts = path.split('/')
            for i, name in enumerate(parts):
                try:
                    parent, obj = obj, obj.__getrpc__(name)
                except (AttributeError, KeyError):
                    raise LookupError('Name "{}" not found in "{}"'.format(
                        name, '/'.join(parts[:i]))) from None

                if not hasattr(obj, '__rpc__'):
                    cls = parent.__class__
                    raise LookupError('"{}.{}.{}" is not exposed'.format(
                        cls.__module__, cls.__qualname__, name))

            self._cache[path] = obj

        return obj

    expose = staticmethod(expose)
    """Alias for :func:`expose()`."""

    def add(self, name):
        """Add the sub-router *name* (stored at ``self.obj.<name>``) to this
        router.

        Convenience wrapper for :meth:`set_sub_router`.

        """
        router = getattr(self.obj, name).router
        self.set_sub_router(router, name)

    def set_sub_router(self, router, name):
        """Set *self* as parent for the *router* named *name*."""
        if type(router) is not Router:
            raise ValueError('"{}" is not a valid router.'.format(router))
        if router.parent is not None:
            raise ValueError('"{}" is already a sub service of "{}"'
                             .format(router.obj, router.parent.obj))
        router.name = name
        router.parent = self


class RpcClient:
    """The RpcClient provides proxy objects for remote calls via its
    :attr:`remote` attribute.

    *channel* is a :class:`~aiomas.channel.Channel` instance for communicating
    with the remote side.

    If *rpc_service* is not ``None``, it will also start its own RPC service so
    the peer can call the functions we provide.

    """
    def __init__(self, channel, rpc_service=None):
        self.__channel = channel
        self.__channel.codec.add_serializer(object, self._serialize_obj,
                                            self._deserialize_obj)
        self.__root_router = None
        self.__service = None
        self.__connection_reset_callback = None

        if rpc_service is not None:
            self.__root_router = rpc_service.router
            self.__service = channel._loop.create_task(
                rpc_service_process(self, rpc_service.router, channel)
            )

    @property
    def channel(self):
        """The communication :class:`~aiomas.channel.Channel` of this instance.
        """
        return self.__channel

    @property
    def service(self):
        """The RPC service process for this connection."""
        return self.__service

    @property
    def remote(self):
        """A :class:`Proxy` for remote methods."""
        return Proxy(self.__channel, '')

    def on_connection_reset(self, callback):
        """Add a *callback* that gets called if the peer closes the connection
        and thus causing the :attr:`service` process to abort.

        *callback* is a callable with a single argument, the exception that the
        :attr:`service` process raises if the connection is reset by the peer.

        If this method is called multiple times, override the current callback
        with the new one.  If *callback* is ``None``, delete the current
        callback.

        Raise a :exc:`ValueError` if *callback* is neither callable nor
        ``None``.

        Raise a :exc:`RuntimeError` if this instance has not service task
        running.

        """
        if self.__service is None:
            raise RuntimeError('This {} instance has no RPC service running.'
                               .format(self.__class__.__name__))

        if not (callback is None or hasattr(callback, '__call__')):
            raise ValueError('{!r} must be "None" or callable but is neither'
                             .format(callback))

        self.__connection_reset_callback = callback

    async def close(self):
        """`Coroutine
        <https://docs.python.org/3/library/asyncio-task.html#coroutine>`_ that
        closes the connection and waits for all sub tasks to finish."""
        await self.__channel.close()
        if self.__service is not None:
            if not self.__service.done():
                self.__service.cancel()
            await self.__service

    def _connection_reset(self, exc):
        """Callback that is executed when a :exc:`ConnectionError` is raised
        within :attr:`service`.

        Using the *connection reset callback* (also see
        :meth:`on_connection_reset()`), the user can check if the connection
        reset was intentional or an error and react accordingly.

        """
        if self.__connection_reset_callback is not None:
            self.__connection_reset_callback(exc)

    def _serialize_obj(self, obj):
        """Fallback serializer for all objects for which no other serializer
        was found.

        It will only serialize valid RPC services and only if "self" has a
        service process running and the RPC service is associated with it.

        Else, raise a :exc:`SerializationError`.

        """
        # Check if "obj" is a valid RPC service

        if not _is_rpc_service(obj):
            raise exceptions.SerializationError(
                'No serializer found for type "{}"'.format(type(obj)))

        # Check if we have a service running:
        if self.__root_router is None:
            raise exceptions.SerializationError(
                'No RPC service running for this side of the connection.  '
                'Cannot send serice proxy.')

        # Check if "obj.router" is self.__root_router or one of its children
        router = obj.router
        while router.parent is not None:
            router = router.parent
        if router is not self.__root_router:
            raise exceptions.SerializationError(
                'RPC service "{!r}" is not exposed for this connection'
                .format(obj))

        return obj.router.path

    def _deserialize_obj(self, path):
        return Proxy(self.__channel, path)


class Proxy:
    """Proxy object for remote objects and functions."""
    def __init__(self, channel, path):
        self._channel = channel
        self._path = path
        self._cache = weakref.WeakValueDictionary()
        self._str = None

    def __repr__(self):
        return '<{}.{} at 0x{:x}>'.format(self.__module__, self, id(self))

    def __str__(self):
        return self._str if self._str else '{}({!r}, {!r})'.format(
            self.__class__.__name__,
            self._channel.get_extra_info('peername')[:2],
            self._path)

    def __getattr__(self, name):
        """Return a new proxy for *name*."""
        if name in self._cache:
            proxy = self._cache[name]
        else:
            path = name if not self._path else '{}/{}'.format(self._path, name)
            proxy = self.__class__(self._channel, path)
            self._cache[name] = proxy

        return proxy

    def __call__(self, *args, **kwargs):
        """Call the remote method represented by this proxy and return its
        result.

        The result is a future, so you need to ``await`` it in order to
        get the actual return value (or exception).

        """
        if not self._path:
            raise AttributeError('No RPC function name specified.')
        return self._channel.send((self._path, args, kwargs))

    def __eq__(self, other):
        return self._channel is other._channel and self._path == other._path

    def __hash__(self):
        return hash(self._channel) + hash(self._path)



from setuptools import setup, find_packages


setup(
    name='aiomas',
    version='2.0.1',
    author='Stefan Scherfke',
    author_email='stefan@sofa-rockers.org',
    description=('Asyncio-based, layered networking library providing '
                 'request-reply channels, RPC, and multi-agent systems.'),
    long_description=(open('README.rst', encoding='utf-8').read() + '\n\n' +
                      open('CHANGES.rst', encoding='utf-8').read() + '\n\n' +
                      open('AUTHORS.rst', encoding='utf-8').read()),
    url='https://aiomas.readthedocs.io',
    python_requires='>=3.5',
    install_requires=[
        'arrow>=0.7',
    ],
    extras_require={
        'mp': ['msgpack-python>=0.4.7'],
        'mpb': ['blosc>=1.3.2', 'msgpack-python>=0.4.7'],
    },
    packages=find_packages(where='src'),
    package_dir={'': 'src'},
    include_package_data=True,
    entry_points={
        'console_scripts': [
        ],
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)



"""
This package imports the codecs that can be used for de- and encoding incoming
and outgoing messages:

- :class:`JSON` uses `JSON <http://www.json.org/>`_
- :class:`MsgPack` uses `msgpack <https://msgpack.org/>`_
- :class:`MsgPackBlosc` uses `msgpack <https://msgpack.org/>`_ and
  `Blosc <http://blosc.org/>`_

All codecs should implement the base class :class:`Codec`.

"""
import inspect
import json
import sys

try:
    import blosc
except ImportError:
    blosc = None
try:
    import msgpack
except ImportError:
    msgpack = None

from .exceptions import SerializationError


__all__ = ['serializable', 'Codec', 'JSON', 'MsgPack', 'MsgPackBlosc']


TYPESIZE = 8 if sys.maxsize > 2**32 else 4


def serializable(cls=None, repr=True):
    """Class decorator that makes the decorated class serializable by
    :mod:`aiomas.codecs`.

    The decorator tries to extract all arguments to the class’ ``__init__()``.
    That means, the arguments must be available as attributes with the same
    name.

    The decorator adds the following methods to the decorated class:

    - ``__asdict__()``: Returns a dict with all __init__ parameters

    - ``__fromdict__(dict)``: Creates a new class instance from *dict*

    - ``__serializer__()``: Returns a tuple with args for
      :meth:`Codec.add_serializer()`

    - ``__repr__()``: Returns a generic instance representation.  Adding this
      method can be deactivated by passing ``repr=False`` to the decorator.

    Example:

    .. code-block:: python

        >>> import aiomas.codecs
        >>>
        >>> @aiomas.codecs.serializable
        ... class A:
        ...     def __init__(self, x, y):
        ...         self.x = x
        ...         self._y = y
        ...
        ...     @property
        ...     def y(self):
        ...         return self._y
        >>>
        >>> codec = aiomas.codecs.JSON()
        >>> codec.add_serializer(*A.__serializer__())
        >>> a = codec.decode(codec.encode(A(1, 2)))
        >>> a
        A(x=1, y=2)

    """
    def wrap(cls):
        attrs = [a for a in inspect.signature(cls).parameters]

        def __asdict__(self):
            return {a: getattr(self, a) for a in attrs}

        @classmethod
        def __fromdict__(cls, attrs):
            return cls(**attrs)

        def __repr__(self):
            args = ('{}={!r}'.format(a, getattr(self, a)) for a in attrs)
            return '{}({})'.format(self.__class__.__name__, ', '.join(args))

        @classmethod
        def __serializer__(cls):
            return (cls, cls.__asdict__, cls.__fromdict__)

        cls.__asdict__ = __asdict__
        cls.__fromdict__ = __fromdict__
        cls.__serializer__ = __serializer__
        if repr:
            cls.__repr__ = __repr__

        return cls

    # The type of "cls" depends on the usage of the decorator.  It's a class if
    # it's used as `@serializable` but ``None`` if used as `@serializable()`.
    if cls is None:
        return wrap
    else:
        return wrap(cls)


class Codec:
    """Base class for all Codecs.

    Subclasses must implement :meth:`encode()` and :meth:`decode()`.

    """
    def __init__(self):
        self._serializers = {}
        self._deserializers = {}

    def __str__(self):
        return '{}[{}]'.format(
            self.__class__.__name__,
            ', '.join(s.__name__ for s in self._serializers),
        )

    def encode(self, data):
        """Encode the given *data* and return a :class:`bytes` object."""
        raise NotImplementedError

    def decode(self, data):
        """Decode *data* from :class:`bytes` to the original data structure."""
        raise NotImplementedError

    def add_serializer(self, type, serialize, deserialize):
        """Add methods to *serialize* and *deserialize* objects typed *type*.

        This can be used to de-/encode objects that the codec otherwise
        couldn't encode.

        *serialize* will receive the unencoded object and needs to return
        an encodable serialization of it.

        *deserialize* will receive an objects representation and should return
        an instance of the original object.

        """
        if type in self._serializers:
            raise ValueError(
                'There is already a serializer for type "{}"'.format(type))
        typeid = len(self._serializers)
        self._serializers[type] = (typeid, serialize)
        self._deserializers[typeid] = deserialize

    def serialize_obj(self, obj):
        """Serialize *obj* to something that the codec can encode."""
        orig_type = otype = type(obj)
        if otype not in self._serializers:
            # Fallback to a generic serializer (if available)
            otype = object

        try:
            typeid, serialize = self._serializers[otype]
        except KeyError:
            raise SerializationError('No serializer found for type "{}"'
                                     .format(orig_type)) from None

        try:
            return {'__type__': (typeid, serialize(obj))}
        except Exception as e:
            raise SerializationError(
                'Could not serialize object "{!r}": {}'.format(obj, e)) from e

    def deserialize_obj(self, obj_repr):
        """Deserialize the original object from *obj_repr*."""
        # This method is called for *all* dicts so we have to check if it
        # contains a desrializable type.
        if '__type__' in obj_repr:
            typeid, data = obj_repr['__type__']
            obj_repr = self._deserializers[typeid](data)
        return obj_repr


class JSON(Codec):
    """A :class:`Codec` that uses *JSON* to encode and decode messages."""

    def encode(self, data):
        return json.dumps(data, default=self.serialize_obj).encode()

    def decode(self, data):
        return json.loads(data.decode(), object_hook=self.deserialize_obj)


class MsgPack(Codec):
    """A :class:`Codec` that uses *msgpack* to encode and decode messages."""
    def __init__(self):
        if msgpack is None:
            msg = (
                'Please install "msgpack-python" to use the {} codec: pip '
                'install -U aiomas[mp]'.format(self.__class__.__name__)
            )
            raise ImportError(msg)
        super().__init__()

    def encode(self, data):
        return msgpack.packb(
            data, default=self.serialize_obj, use_bin_type=True)

    def decode(self, data):
        return msgpack.unpackb(data,
                               object_hook=self.deserialize_obj,
                               use_list=False,
                               encoding='utf-8')


class MsgPackBlosc(Codec):
    """A :class:`Codec` that uses *msgpack* to encode and decode messages and
    *blosc* to compress them."""
    def __init__(self):
        if msgpack is None or blosc is None:
            msg = (
                'Please install "msgpack-python" and "blosc" to use the {} '
                'codec: pip install -U aiomas[mpb]'
                .format(self.__class__.__name__)
            )
            raise ImportError(msg)
        super().__init__()

    def encode(self, data):
        return blosc.compress(msgpack.packb(
            data, default=self.serialize_obj, use_bin_type=True), TYPESIZE)

    def decode(self, data):
        return msgpack.unpackb(blosc.decompress(bytes(data)),
                               object_hook=self.deserialize_obj,
                               use_list=False,
                               encoding='utf-8')
